"""
API for TensorFlow usage.
"""

from typing import Any, Callable, Optional, Sequence, Tuple

import numpy as np  # type: ignore
import tensorflow as tf  # type: ignore

from targetran._functional import (
    _tf_convert,
    _tf_round_to_int,
    _tf_resize_image,
    _tf_pad_image,
    _tf_gather_image
)


from targetran._transform import (
    _flip_left_right,
    _flip_up_down,
    _rotate_90,
    _rotate_90_and_pad,
    _rotate,
    _shear,
    _crop,
    _resize,
    _translate,
    _get_random_crop_inputs,
    _get_random_size_fractions
)


def np_to_tf(
        image_seq: Sequence[np.ndarray],
        bboxes_seq: Sequence[np.ndarray],
        labels_seq: Sequence[np.ndarray]
) -> Tuple[Sequence[tf.Tensor], Sequence[tf.Tensor], Sequence[tf.Tensor]]:
    """
    Convert Numpy array seqs to TF (eager) tensor seqs.
    """
    tuples = [
        (_tf_convert(image),
         tf.reshape(_tf_convert(bboxes), (-1, 4)),
         _tf_convert(labels))
        for image, bboxes, labels in zip(image_seq, bboxes_seq, labels_seq)
    ]
    tf_image_seq, tf_bboxes_seq, tf_labels_seq = tuple(zip(*tuples))
    return tf_image_seq, tf_bboxes_seq, tf_labels_seq


def seqs_to_tf_dataset(
        image_seq: Sequence[np.ndarray],
        bboxes_seq: Sequence[np.ndarray],
        labels_seq: Sequence[np.ndarray]
) -> tf.data.Dataset:

    tf_image_seq, tf_bboxes_seq, tf_labels_seq = np_to_tf(
        image_seq, bboxes_seq, labels_seq
    )

    # Tensors of different shapes can be included in a TF Dataset
    # as ragged-tensors.
    ds = tf.data.Dataset.zip((
        tf.data.Dataset.from_tensor_slices(tf.ragged.stack(tf_image_seq)),
        tf.data.Dataset.from_tensor_slices(tf.ragged.stack(tf_bboxes_seq)),
        tf.data.Dataset.from_tensor_slices(tf.ragged.stack(tf_labels_seq))
    ))
    # However, our transformations expect normal tensors, so the ragged-tensors
    # have to be first converted back to tensors during mapping. Therefore,
    # the whole point of using ragged-tensors is ONLY for building a Dataset...
    # Note that the label ragged-tensors are of rank-0, so they are implicitly
    # converted to tensors during mapping. Strange TF Dataset behaviour...
    ds = ds.map(lambda i, b, l: (i.to_tensor(), b.to_tensor(), l))
    return ds


def tf_flip_left_right(
        image: tf.Tensor,
        bboxes: tf.Tensor,
        labels: tf.Tensor
) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
    return _flip_left_right(
        image, bboxes, labels,
        _tf_convert, tf.shape, tf.reshape, tf.concat
    )


def tf_flip_up_down(
        image: tf.Tensor,
        bboxes: tf.Tensor,
        labels: tf.Tensor
) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
    return _flip_up_down(
        image, bboxes, labels,
        _tf_convert, tf.shape, tf.reshape, tf.concat
    )


def tf_resize(
        image: tf.Tensor,
        bboxes: tf.Tensor,
        labels: tf.Tensor,
        dest_size: Tuple[int, int]
) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
    return _resize(
        image, bboxes, labels, dest_size,
        _tf_convert, tf.shape, tf.reshape, _tf_resize_image, tf.concat
    )


def tf_rotate_90(
        image: tf.Tensor,
        bboxes: tf.Tensor,
        labels: tf.Tensor
) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
    return _rotate_90(
        image, bboxes, labels,
        _tf_convert, tf.shape, tf.reshape, tf.transpose, tf.concat
    )


def tf_rotate_90_and_pad(
        image: tf.Tensor,
        bboxes: tf.Tensor,
        labels: tf.Tensor
) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
    return _rotate_90_and_pad(
        image, bboxes, labels,
        _tf_convert, tf.shape, tf.reshape, tf.transpose, tf.concat,
        tf.where, tf.math.ceil, tf.math.floor, _tf_pad_image
    )


def tf_rotate(
        image: tf.Tensor,
        bboxes: tf.Tensor,
        labels: tf.Tensor,
        angle_deg: float
) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
    return _rotate(
        image, bboxes, labels, angle_deg,
        _tf_convert, tf.shape, tf.reshape, tf.expand_dims, tf.squeeze,
        _tf_pad_image, tf.range, _tf_round_to_int, tf.repeat, tf.tile,
        tf.stack, tf.concat, tf.cos, tf.sin, tf.matmul, tf.clip_by_value,
        _tf_gather_image, tf.identity, tf.reduce_max, tf.reduce_min,
        tf.logical_and, tf.boolean_mask
    )


def tf_shear(
        image: tf.Tensor,
        bboxes: tf.Tensor,
        labels: tf.Tensor,
        angle_deg: float
) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
    return _shear(
        image, bboxes, labels, angle_deg,
        _tf_convert, tf.shape, tf.reshape, tf.expand_dims, tf.squeeze,
        _tf_pad_image, tf.range, _tf_round_to_int, tf.repeat, tf.tile,
        tf.stack, tf.concat, tf.tan, tf.matmul, tf.clip_by_value,
        _tf_gather_image, tf.identity, tf.reduce_max, tf.reduce_min,
        tf.logical_and, tf.boolean_mask
    )


def _tf_get_random_crop_inputs(
        image_height: int,
        image_width: int,
        height_fraction_range: Tuple[float, float],
        width_fraction_range: Tuple[float, float],
        rand_fn: Callable[..., tf.Tensor]
) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor, tf.Tensor]:
    return _get_random_crop_inputs(
        image_height, image_width, height_fraction_range, width_fraction_range,
        rand_fn, _tf_convert, _tf_round_to_int
    )


def tf_crop(
        image: tf.Tensor,
        bboxes: tf.Tensor,
        labels: tf.Tensor,
        offset_height: int,
        offset_width: int,
        cropped_image_height: int,
        cropped_image_width: int
) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
    return _crop(
        image, bboxes, labels,
        offset_height, offset_width,
        cropped_image_height, cropped_image_width,
        _tf_convert, tf.shape, tf.reshape, tf.concat,
        tf.logical_and, tf.squeeze, tf.clip_by_value, tf.boolean_mask
    )


def tf_translate(
        image: tf.Tensor,
        bboxes: tf.Tensor,
        labels: tf.Tensor,
        translate_height: int,
        translate_width: int
) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
    return _translate(
        image, bboxes, labels,
        translate_height, translate_width,
        _tf_convert, tf.shape, tf.reshape, tf.where, tf.abs, tf.concat,
        tf.logical_and, tf.squeeze, tf.clip_by_value, tf.boolean_mask,
        _tf_pad_image
    )


class TFResize:

    def __init__(self, dest_size: Tuple[int, int]) -> None:
        self.dest_size = dest_size

    def __call__(
            self,
            image: tf.Tensor,
            bboxes: tf.Tensor,
            labels: tf.Tensor
    ) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
        return tf_resize(image, bboxes, labels, self.dest_size)


class TFRandomTransform:

    def __init__(
            self,
            tf_fn: Callable[..., Tuple[tf.Tensor, tf.Tensor, tf.Tensor]],
            probability: float,
            seed: Optional[int],
    ) -> None:
        self._tf_fn = tf_fn
        self.probability = probability
        self._rng = tf.random.Generator.from_seed(seed) if seed is not None \
            else tf.random.Generator.from_non_deterministic_state()
        self._rand_fn: Callable[..., tf.Tensor] = lambda: self._rng.uniform([])

    def __call__(
            self,
            image: tf.Tensor,
            bboxes: tf.Tensor,
            labels: tf.Tensor,
            *args: Any,
            **kwargs: Any
    ) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:

        image = _tf_convert(image)
        bboxes = _tf_convert(bboxes)
        labels = _tf_convert(labels)

        if self._rand_fn() < self.probability:
            return self._tf_fn(image, bboxes, labels, *args, **kwargs)
        return image, bboxes, labels


class TFRandomFlipLeftRight(TFRandomTransform):

    def __init__(
            self,
            probability: float = 0.7,
            seed: Optional[int] = None
    ) -> None:
        super().__init__(tf_flip_left_right, probability, seed)

    def __call__(
            self,
            image: tf.Tensor,
            bboxes: tf.Tensor,
            labels: tf.Tensor,
            *args: Any,
            **kwargs: Any
    ) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
        return super().__call__(image, bboxes, labels)


class TFRandomFlipUpDown(TFRandomTransform):

    def __init__(
            self,
            probability: float = 0.7,
            seed: Optional[int] = None
    ) -> None:
        super().__init__(tf_flip_up_down, probability, seed)

    def __call__(
            self,
            image: tf.Tensor,
            bboxes: tf.Tensor,
            labels: tf.Tensor,
            *args: Any,
            **kwargs: Any
    ) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
        return super().__call__(image, bboxes, labels)


class TFRandomRotate90(TFRandomTransform):

    def __init__(
            self,
            probability: float = 0.7,
            seed: Optional[int] = None
    ) -> None:
        super().__init__(tf_rotate_90, probability, seed)

    def __call__(
            self,
            image: tf.Tensor,
            bboxes: tf.Tensor,
            labels: tf.Tensor,
            *args: Any,
            **kwargs: Any
    ) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
        return super().__call__(image, bboxes, labels)


class TFRandomRotate90AndPad(TFRandomTransform):

    def __init__(
            self,
            probability: float = 0.7,
            seed: Optional[int] = None
    ) -> None:
        super().__init__(tf_rotate_90_and_pad, probability, seed)

    def __call__(
            self,
            image: tf.Tensor,
            bboxes: tf.Tensor,
            labels: tf.Tensor,
            *args: Any,
            **kwargs: Any
    ) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
        return super().__call__(image, bboxes, labels)


class TFRandomRotate(TFRandomTransform):

    def __init__(
            self,
            angle_deg_range: Tuple[float, float] = (-15.0, 15.0),
            probability: float = 0.7,
            seed: Optional[int] = None
    ) -> None:
        super().__init__(tf_rotate, probability, seed)
        assert angle_deg_range[0] < angle_deg_range[1]
        self.angle_deg_range = angle_deg_range

    def __call__(
            self,
            image: tf.Tensor,
            bboxes: tf.Tensor,
            labels: tf.Tensor,
            *args: Any,
            **kwargs: Any
    ) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:

        angle_deg = \
            _tf_convert(self.angle_deg_range[1] - self.angle_deg_range[0]) \
            * self._rand_fn() + _tf_convert(self.angle_deg_range[0])

        return super().__call__(image, bboxes, labels, angle_deg)


class TFRandomShear(TFRandomTransform):

    def __init__(
            self,
            angle_deg_range: Tuple[float, float] = (-15.0, 15.0),
            probability: float = 0.7,
            seed: Optional[int] = None
    ) -> None:
        super().__init__(tf_shear, probability, seed)
        assert -90.0 < angle_deg_range[0] < angle_deg_range[1] < 90.0
        self.angle_deg_range = angle_deg_range

    def __call__(
            self,
            image: tf.Tensor,
            bboxes: tf.Tensor,
            labels: tf.Tensor,
            *args: Any,
            **kwargs: Any
    ) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:

        angle_deg = \
            _tf_convert(self.angle_deg_range[1] - self.angle_deg_range[0]) \
            * self._rand_fn() + _tf_convert(self.angle_deg_range[0])

        return super().__call__(image, bboxes, labels, angle_deg)


class TFRandomCrop(TFRandomTransform):

    def __init__(
            self,
            crop_height_fraction_range: Tuple[float, float] = (0.7, 0.9),
            crop_width_fraction_range: Tuple[float, float] = (0.7, 0.9),
            probability: float = 0.7,
            seed: Optional[int] = None
    ) -> None:
        super().__init__(tf_crop, probability, seed)
        self.crop_height_fraction_range = crop_height_fraction_range
        self.crop_width_fraction_range = crop_width_fraction_range

    def __call__(
            self,
            image: tf.Tensor,
            bboxes: tf.Tensor,
            labels: tf.Tensor,
            *args: Any,
            **kwargs: Any
    ) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:

        offset_height, offset_width, cropped_height, cropped_width = \
            _tf_get_random_crop_inputs(
                tf.shape(image)[0], tf.shape(image)[1],
                self.crop_height_fraction_range,
                self.crop_width_fraction_range,
                self._rand_fn
            )

        return super().__call__(
            image, bboxes, labels,
            offset_height, offset_width, cropped_height, cropped_width
        )


class TFRandomTranslate(TFRandomTransform):

    def __init__(
            self,
            translate_height_fraction_range: Tuple[float, float] = (-0.2, 0.2),
            translate_width_fraction_range: Tuple[float, float] = (-0.2, 0.2),
            probability: float = 0.7,
            seed: Optional[int] = None
    ) -> None:
        super().__init__(tf_translate, probability, seed)
        self.translate_height_fraction_range = translate_height_fraction_range
        self.translate_width_fraction_range = translate_width_fraction_range

    def __call__(
            self,
            image: tf.Tensor,
            bboxes: tf.Tensor,
            labels: tf.Tensor,
            *args: Any,
            **kwargs: Any
    ) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:

        height_fraction, width_fraction = _get_random_size_fractions(
            self.translate_height_fraction_range,
            self.translate_width_fraction_range,
            self._rand_fn, _tf_convert
        )

        translate_height = _tf_round_to_int(
            _tf_convert(tf.shape(image)[0]) * height_fraction
        )
        translate_width = _tf_round_to_int(
            _tf_convert(tf.shape(image)[1]) * width_fraction
        )

        return super().__call__(
            image, bboxes, labels, translate_height, translate_width
        )



from ._tf import (
    np_to_tf,
    seqs_to_tf_dataset,
    tf_flip_left_right,
    tf_flip_up_down,
    tf_rotate_90,
    tf_rotate_90_and_pad,
    tf_rotate,
    tf_crop,
    tf_translate,
    tf_shear,
    TFResize,
    TFRandomTransform,
    TFRandomFlipLeftRight,
    TFRandomFlipUpDown,
    TFRandomRotate90,
    TFRandomRotate90AndPad,
    TFRandomRotate,
    TFRandomCrop,
    TFRandomTranslate,
    TFRandomShear
)



"""
API for Numpy usage.
"""

from typing import Any, Callable, Optional, Tuple

import numpy as np  # type: ignore

from targetran._functional import (
    _np_convert,
    _np_range,
    _np_round_to_int,
    _np_resize_image,
    _np_boolean_mask,
    _np_logical_and,
    _np_pad_image,
    _np_gather_image
)

from targetran._transform import (
    _flip_left_right,
    _flip_up_down,
    _rotate_90,
    _rotate_90_and_pad,
    _rotate,
    _shear,
    _crop,
    _resize,
    _translate,
    _get_random_crop_inputs,
    _get_random_size_fractions
)


def flip_left_right(
        image: np.ndarray,
        bboxes: np.ndarray,
        labels: np.ndarray
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    return _flip_left_right(
        image, bboxes, labels,
        _np_convert, np.shape, np.reshape, np.concatenate
    )


def flip_up_down(
        image: np.ndarray,
        bboxes: np.ndarray,
        labels: np.ndarray
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    return _flip_up_down(
        image, bboxes, labels,
        _np_convert, np.shape, np.reshape, np.concatenate
    )


def resize(
        image: np.ndarray,
        bboxes: np.ndarray,
        labels: np.ndarray,
        dest_size: Tuple[int, int]
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    return _resize(
        image, bboxes, labels, dest_size,
        _np_convert, np.shape, np.reshape, _np_resize_image, np.concatenate
    )


def rotate_90(
        image: np.ndarray,
        bboxes: np.ndarray,
        labels: np.ndarray
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    return _rotate_90(
        image, bboxes, labels,
        _np_convert, np.shape, np.reshape, np.transpose, np.concatenate
    )


def rotate_90_and_pad(
        image: np.ndarray,
        bboxes: np.ndarray,
        labels: np.ndarray
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    return _rotate_90_and_pad(
        image, bboxes, labels,
        _np_convert, np.shape, np.reshape, np.transpose, np.concatenate,
        np.where, np.ceil, np.floor, _np_pad_image
    )


def rotate(
        image: np.ndarray,
        bboxes: np.ndarray,
        labels: np.ndarray,
        angle_deg: float
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    return _rotate(
        image, bboxes, labels, angle_deg,
        _np_convert, np.shape, np.reshape, np.expand_dims, np.squeeze,
        _np_pad_image, _np_range, _np_round_to_int, np.repeat, np.tile,
        np.stack, np.concatenate, np.cos, np.sin, np.matmul, np.clip,
        _np_gather_image, np.copy, np.max, np.min,
        _np_logical_and, _np_boolean_mask
    )


def shear(
        image: np.ndarray,
        bboxes: np.ndarray,
        labels: np.ndarray,
        angle_deg: float
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    return _shear(
        image, bboxes, labels, angle_deg,
        _np_convert, np.shape, np.reshape, np.expand_dims, np.squeeze,
        _np_pad_image, _np_range, _np_round_to_int, np.repeat, np.tile,
        np.stack, np.concatenate, np.tan, np.matmul, np.clip,
        _np_gather_image, np.copy, np.max, np.min,
        _np_logical_and, _np_boolean_mask
    )


def _np_get_random_crop_inputs(
        image_height: int,
        image_width: int,
        height_fraction_range: Tuple[float, float],
        width_fraction_range: Tuple[float, float],
        rand_fn: Callable[..., np.ndarray]
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    return _get_random_crop_inputs(
        image_height, image_width, height_fraction_range, width_fraction_range,
        rand_fn, _np_convert, _np_round_to_int
    )


def crop(
        image: np.ndarray,
        bboxes: np.ndarray,
        labels: np.ndarray,
        offset_height: int,
        offset_width: int,
        cropped_image_height: int,
        cropped_image_width: int
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    return _crop(
        image, bboxes, labels,
        offset_height, offset_width,
        cropped_image_height, cropped_image_width,
        _np_convert, np.shape, np.reshape, np.concatenate,
        _np_logical_and, np.squeeze, np.clip, _np_boolean_mask
    )


def translate(
        image: np.ndarray,
        bboxes: np.ndarray,
        labels: np.ndarray,
        translate_height: int,
        translate_width: int
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    return _translate(
        image, bboxes, labels,
        translate_height, translate_width,
        _np_convert, np.shape, np.reshape, np.where, np.abs, np.concatenate,
        _np_logical_and, np.squeeze, np.clip, _np_boolean_mask, _np_pad_image
    )


class Resize:

    def __init__(self, dest_size: Tuple[int, int]) -> None:
        self.dest_size = dest_size

    def __call__(
            self,
            image: np.ndarray,
            bboxes: np.ndarray,
            labels: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        return resize(image, bboxes, labels, self.dest_size)


class RandomTransform:

    def __init__(
            self,
            np_fn: Callable[..., Tuple[np.ndarray, np.ndarray, np.ndarray]],
            probability: float,
            seed: Optional[int],
    ) -> None:
        self._np_fn = np_fn
        self.probability = probability
        self._rng = np.random.default_rng(seed=seed)
        self._rand_fn: Callable[..., np.ndarray] = lambda: self._rng.random()

    def __call__(
            self,
            image: np.ndarray,
            bboxes: np.ndarray,
            labels: np.ndarray,
            *args: Any,
            **kwargs: Any
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:

        image = _np_convert(image)
        bboxes = _np_convert(bboxes)
        labels = _np_convert(labels)

        if self._rand_fn() < self.probability:
            return self._np_fn(image, bboxes, labels, *args, **kwargs)
        return image, bboxes, labels


class RandomFlipLeftRight(RandomTransform):

    def __init__(
            self,
            probability: float = 0.7,
            seed: Optional[int] = None
    ) -> None:
        super().__init__(flip_left_right, probability, seed)

    def __call__(
            self,
            image: np.ndarray,
            bboxes: np.ndarray,
            labels: np.ndarray,
            *args: Any,
            **kwargs: Any
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        return super().__call__(image, bboxes, labels)


class RandomFlipUpDown(RandomTransform):

    def __init__(
            self,
            probability: float = 0.7,
            seed: Optional[int] = None
    ) -> None:
        super().__init__(flip_up_down, probability, seed)

    def __call__(
            self,
            image: np.ndarray,
            bboxes: np.ndarray,
            labels: np.ndarray,
            *args: Any,
            **kwargs: Any
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        return super().__call__(image, bboxes, labels)


class RandomRotate90(RandomTransform):

    def __init__(
            self,
            probability: float = 0.7,
            seed: Optional[int] = None
    ) -> None:
        super().__init__(rotate_90, probability, seed)

    def __call__(
            self,
            image: np.ndarray,
            bboxes: np.ndarray,
            labels: np.ndarray,
            *args: Any,
            **kwargs: Any
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        return super().__call__(image, bboxes, labels)


class RandomRotate90AndPad(RandomTransform):

    def __init__(
            self,
            probability: float = 0.7,
            seed: Optional[int] = None
    ) -> None:
        super().__init__(rotate_90_and_pad, probability, seed)

    def __call__(
            self,
            image: np.ndarray,
            bboxes: np.ndarray,
            labels: np.ndarray,
            *args: Any,
            **kwargs: Any
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        return super().__call__(image, bboxes, labels)


class RandomRotate(RandomTransform):

    def __init__(
            self,
            angle_deg_range: Tuple[float, float] = (-15.0, 15.0),
            probability: float = 0.7,
            seed: Optional[int] = None
    ) -> None:
        super().__init__(rotate, probability, seed)
        assert angle_deg_range[0] < angle_deg_range[1]
        self.angle_deg_range = angle_deg_range

    def __call__(
            self,
            image: np.ndarray,
            bboxes: np.ndarray,
            labels: np.ndarray,
            *args: Any,
            **kwargs: Any
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:

        angle_deg = \
            self.angle_deg_range[1] - self.angle_deg_range[0] \
            * self._rand_fn() + self.angle_deg_range[0]

        return super().__call__(image, bboxes, labels, angle_deg)


class RandomShear(RandomTransform):

    def __init__(
            self,
            angle_deg_range: Tuple[float, float] = (-15.0, 15.0),
            probability: float = 0.7,
            seed: Optional[int] = None
    ) -> None:
        super().__init__(shear, probability, seed)
        assert -90.0 < angle_deg_range[0] < angle_deg_range[1] < 90.0
        self.angle_deg_range = angle_deg_range

    def __call__(
            self,
            image: np.ndarray,
            bboxes: np.ndarray,
            labels: np.ndarray,
            *args: Any,
            **kwargs: Any
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:

        angle_deg = \
            self.angle_deg_range[1] - self.angle_deg_range[0] \
            * self._rand_fn() + self.angle_deg_range[0]

        return super().__call__(image, bboxes, labels, angle_deg)


class RandomCrop(RandomTransform):

    def __init__(
            self,
            crop_height_fraction_range: Tuple[float, float] = (0.7, 0.9),
            crop_width_fraction_range: Tuple[float, float] = (0.7, 0.9),
            probability: float = 0.7,
            seed: Optional[int] = None
    ) -> None:
        super().__init__(crop, probability, seed)
        self.crop_height_fraction_range = crop_height_fraction_range
        self.crop_width_fraction_range = crop_width_fraction_range

    def __call__(
            self,
            image: np.ndarray,
            bboxes: np.ndarray,
            labels: np.ndarray,
            *args: Any,
            **kwargs: Any
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:

        offset_height, offset_width, cropped_height, cropped_width = \
            _np_get_random_crop_inputs(
                np.shape(image)[0], np.shape(image)[1],
                self.crop_height_fraction_range,
                self.crop_width_fraction_range,
                self._rand_fn
            )

        return super().__call__(
            image, bboxes, labels,
            offset_height, offset_width, cropped_height, cropped_width
        )


class RandomTranslate(RandomTransform):

    def __init__(
            self,
            translate_height_fraction_range: Tuple[float, float] = (-0.2, 0.2),
            translate_width_fraction_range: Tuple[float, float] = (-0.2, 0.2),
            probability: float = 0.7,
            seed: Optional[int] = None
    ) -> None:
        super().__init__(translate, probability, seed)
        self.translate_height_fraction_range = translate_height_fraction_range
        self.translate_width_fraction_range = translate_width_fraction_range

    def __call__(
            self,
            image: np.ndarray,
            bboxes: np.ndarray,
            labels: np.ndarray,
            *args: Any,
            **kwargs: Any
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:

        height_fraction, width_fraction = _get_random_size_fractions(
            self.translate_height_fraction_range,
            self.translate_width_fraction_range,
            self._rand_fn, _np_convert
        )

        translate_height = _np_round_to_int(
            np.shape(image)[0] * height_fraction
        )
        translate_width = _np_round_to_int(
            np.shape(image)[1] * width_fraction
        )

        return super().__call__(
            image, bboxes, labels, translate_height, translate_width
        )



"""
Image and target transform utilities.
"""

from typing import Callable, List, Tuple, TypeVar

import numpy as np  # type: ignore
import tensorflow as tf  # type: ignore


T = TypeVar("T", np.ndarray, tf.Tensor)


def _sanitise(
        image: T,
        bboxes: T,
        labels: T,
        convert_fn: Callable[..., T],
        reshape_fn: Callable[[T, Tuple[int, ...]], T],
) -> Tuple[T, T, T]:
    """
    Try to convert inputs to the expected format.
    """
    image = convert_fn(image)
    bboxes = reshape_fn(convert_fn(bboxes), (-1, 4))
    labels = convert_fn(labels)
    return image, bboxes, labels


def _flip_left_right(
        image: T,
        bboxes: T,
        labels: T,
        convert_fn: Callable[..., T],
        shape_fn: Callable[[T], Tuple[int, ...]],
        reshape_fn: Callable[[T, Tuple[int, ...]], T],
        concat_fn: Callable[[List[T], int], T]
) -> Tuple[T, T, T]:
    """
    image: [h, w, c]
    bboxes: [[top_left_x, top_left_y, width, height], ...]
    labels: [0, 1, 0, ...]
    """
    image, bboxes, labels = _sanitise(
        image, bboxes, labels, convert_fn, reshape_fn
    )

    image_shape = shape_fn(image)
    assert len(image_shape) == 3

    image_width = convert_fn(image_shape[1])

    image = image[:, ::-1, :]

    xs = image_width - bboxes[:, :1] - bboxes[:, 2:3]
    bboxes = concat_fn([xs, bboxes[:, 1:]], 1)

    return image, bboxes, labels


def _flip_up_down(
        image: T,
        bboxes: T,
        labels: T,
        convert_fn: Callable[..., T],
        shape_fn: Callable[[T], Tuple[int, ...]],
        reshape_fn: Callable[[T, Tuple[int, ...]], T],
        concat_fn: Callable[[List[T], int], T]
) -> Tuple[T, T, T]:
    """
    image: [h, w, c]
    bboxes: [[top_left_x, top_left_y, width, height], ...]
    labels: [0, 1, 0, ...]
    """
    image, bboxes, labels = _sanitise(
        image, bboxes, labels, convert_fn, reshape_fn
    )

    image_shape = shape_fn(image)
    assert len(image_shape) == 3

    image_height = convert_fn(image_shape[0])

    image = image[::-1, :, :]

    ys = image_height - bboxes[:, 1:2] - bboxes[:, 3:]
    bboxes = concat_fn([bboxes[:, :1], ys, bboxes[:, 2:]], 1)

    return image, bboxes, labels


def _rotate_90(
        image: T,
        bboxes: T,
        labels: T,
        convert_fn: Callable[..., T],
        shape_fn: Callable[[T], Tuple[int, ...]],
        reshape_fn: Callable[[T, Tuple[int, ...]], T],
        transpose_fn: Callable[[T, Tuple[int, ...]], T],
        concat_fn: Callable[[List[T], int], T]
) -> Tuple[T, T, T]:
    """
    Rotate 90 degrees anti-clockwise.

    image: [h, w, c]
    bboxes: [[top_left_x, top_left_y, width, height], ...]
    labels: [0, 1, 0, ...]
    """
    image, bboxes, labels = _sanitise(
        image, bboxes, labels, convert_fn, reshape_fn
    )

    image_shape = shape_fn(image)
    assert len(image_shape) == 3

    image_width = convert_fn(image_shape[1])

    image = transpose_fn(image, (1, 0, 2))[::-1, :, :]

    bboxes = concat_fn([
        bboxes[:, 1:2],
        image_width - bboxes[:, :1] - bboxes[:, 2:3],
        bboxes[:, 3:],
        bboxes[:, 2:3],
    ], 1)

    return image, bboxes, labels


def _translate_bboxes(
        bboxes: T,
        top_offset: T,
        left_offset: T,
        concat_fn: Callable[[List[T], int], T]
) -> T:
    return concat_fn([
        bboxes[:, :1] + left_offset,
        bboxes[:, 1:2] + top_offset,
        bboxes[:, 2:3],
        bboxes[:, 3:],
    ], 1)


def _pad(
        image: T,
        bboxes: T,
        labels: T,
        pad_offsets: T,
        shape_fn: Callable[[T], Tuple[int, ...]],
        reshape_fn: Callable[[T, Tuple[int, ...]], T],
        pad_image_fn: Callable[[T, T], T],
        concat_fn: Callable[[List[T], int], T]
) -> Tuple[T, T, T]:
    """
    image: [h, w, c]
    bboxes: [[top_left_x, top_left_y, width, height], ...]
    labels: [0, 1, 0, ...]
    pad_offsets: (top, bottom, left, right)
    """
    image_shape = shape_fn(image)
    assert len(image_shape) == 3
    bboxes = reshape_fn(bboxes, (-1, 4))

    image = pad_image_fn(image, pad_offsets)

    bboxes = _translate_bboxes(
        bboxes, pad_offsets[0], pad_offsets[2], concat_fn
    )

    return image, bboxes, labels


def _rotate_90_and_pad(
        image: T,
        bboxes: T,
        labels: T,
        convert_fn: Callable[..., T],
        shape_fn: Callable[[T], Tuple[int, ...]],
        reshape_fn: Callable[[T, Tuple[int, ...]], T],
        transpose_fn: Callable[[T, Tuple[int, ...]], T],
        concat_fn: Callable[[List[T], int], T],
        where_fn: Callable[[T, T, T], T],
        ceil_fn: Callable[[T], T],
        floor_fn: Callable[[T], T],
        pad_image_fn: Callable[[T, T], T]
) -> Tuple[T, T, T]:
    """
    Rotate 90 degrees anti-clockwise and *try* to pad to the same aspect ratio.

    image: [h, w, c]
    bboxes: [[top_left_x, top_left_y, width, height], ...]
    labels: [0, 1, 0, ...]
    """
    image, bboxes, labels = _sanitise(
        image, bboxes, labels, convert_fn, reshape_fn
    )

    image, bboxes, labels = _rotate_90(
        image, bboxes, labels,
        convert_fn, shape_fn, reshape_fn, transpose_fn, concat_fn
    )
    new_height = convert_fn(shape_fn(image)[0])
    new_width = convert_fn(shape_fn(image)[1])
    longer = where_fn(new_height > new_width, new_height, new_width)
    shorter = where_fn(new_height < new_width, new_height, new_width)

    pad_length = longer ** 2 / shorter - shorter
    half_pad_length = pad_length / convert_fn(2.0)
    pad_major = int(ceil_fn(half_pad_length))
    pad_minor = int(floor_fn(half_pad_length))

    pad_offsets = where_fn(
        new_height > new_width,
        convert_fn((0, 0, pad_major, pad_minor)),
        convert_fn((pad_major, pad_minor, 0, 0))
    )
    return _pad(
        image, bboxes, labels,
        pad_offsets, shape_fn, reshape_fn, pad_image_fn, concat_fn
    )


def _affine_transform(
        image: T,
        bboxes: T,
        labels: T,
        convert_fn: Callable[..., T],
        shape_fn: Callable[[T], Tuple[int, ...]],
        reshape_fn: Callable[[T, Tuple[int, ...]], T],
        expand_dim_fn: Callable[[T, int], T],
        squeeze_fn: Callable[[T, int], T],
        pad_image_fn: Callable[[T, T], T],
        range_fn: Callable[[int, int, int], T],
        round_to_int_fn: Callable[[T], T],
        repeat_fn: Callable[[T, T], T],
        tile_fn: Callable[[T, T], T],
        stack_fn: Callable[[List[T], int], T],
        concat_fn: Callable[[List[T], int], T],
        image_dest_tran_mat: T,
        bboxes_tran_mat: T,
        matmul_fn: Callable[[T, T], T],
        clip_fn: Callable[[T, T, T], T],
        gather_image_fn: Callable[[T, T], T],
        copy_fn: Callable[[T], T],
        max_fn: Callable[[T, int], T],
        min_fn: Callable[[T, int], T],
        logical_and_fn: Callable[[T, T], T],
        boolean_mask_fn: Callable[[T, T], T]
) -> Tuple[T, T, T]:
    """
    image: [h, w, c]
    bboxes: [[top_left_x, top_left_y, width, height], ...]
    labels: [0, 1, 0, ...]
    """
    image, bboxes, labels = _sanitise(
        image, bboxes, labels, convert_fn, reshape_fn
    )

    image_shape = shape_fn(image)
    assert len(image_shape) == 3

    height, width = int(image_shape[0]), int(image_shape[1])
    h_mod, w_mod = height % 2, width % 2
    num_channels = int(image_shape[2])

    # Pad image to provide a zero-value pixel frame for clipping use below.
    pad_offsets = convert_fn([1, 1, 1, 1])
    image = pad_image_fn(image, pad_offsets)

    # References:
    # https://www.kaggle.com/cdeotte/rotation-augmentation-gpu-tpu-0-96

    # Destination indices. Note that (-foo // 2) != -(foo // 2).
    row_idxes = repeat_fn(  # Along y-axis, from top to bottom.
        range_fn(-(height // 2) + 1 - h_mod, height // 2 + 1, 1),
        round_to_int_fn(convert_fn([width]))
    )
    col_idxes = tile_fn(  # Along x-axis, from left to right.
        range_fn(-(width // 2) + 1 - w_mod, width // 2 + 1, 1),
        round_to_int_fn(convert_fn([height]))
    )
    # Note the (col, row) -> (x, y) swapping.
    image_dest_idxes = stack_fn([col_idxes, row_idxes], 0)

    # Transform image, with clipping.
    new_image_dest_idxes = matmul_fn(
        image_dest_tran_mat, convert_fn(image_dest_idxes)
    )
    clipped_new_image_dest_idxes = clip_fn(
        new_image_dest_idxes,
        # Note the extra idx for the padded frame.
        convert_fn([
            [-(width // 2) - w_mod], [-(height // 2) - h_mod]
        ]),
        convert_fn([
            [width // 2 + 1], [height // 2 + 1]
        ])
    )

    # Assigning original pixel values to new positions.
    image_orig_idxes = concat_fn([
        # Rows.
        clipped_new_image_dest_idxes[1:] + convert_fn(height // 2 + h_mod),
        # Columns.
        clipped_new_image_dest_idxes[:1] + convert_fn(width // 2 + w_mod)
    ], 0)
    image_orig_idxes = round_to_int_fn(image_orig_idxes)
    values = gather_image_fn(image, image_orig_idxes)
    new_image = reshape_fn(values, (height, width, num_channels))

    # Transform bboxes.
    top_left_xs = bboxes[:, :1]
    top_left_ys = bboxes[:, 1:2]
    top_right_xs = bboxes[:, :1] + bboxes[:, 2:3] - 1
    top_right_ys = bboxes[:, 1:2]
    bottom_left_xs = copy_fn(top_left_xs)
    bottom_left_ys = copy_fn(top_left_ys + bboxes[:, 3:] - 1)
    bottom_right_xs = copy_fn(top_right_xs)
    bottom_right_ys = copy_fn(top_right_ys + bboxes[:, 3:] - 1)

    xs = concat_fn(
        [top_left_xs - convert_fn(width // 2 - 1 + w_mod),
         top_right_xs - convert_fn(width // 2 - 1 + w_mod),
         bottom_left_xs - convert_fn(width // 2 - 1 + w_mod),
         bottom_right_xs - convert_fn(width // 2 - 1 + w_mod)],
        1
    )
    ys = concat_fn(
        [top_left_ys - convert_fn(height // 2 - 1 + h_mod),
         top_right_ys - convert_fn(height // 2 - 1 + h_mod),
         bottom_left_ys - convert_fn(height // 2 - 1 + h_mod),
         bottom_right_ys - convert_fn(height // 2 - 1 + h_mod)],
        1
    )
    bboxes_idxes = stack_fn([xs, ys], 1)  # Shape: [num_bboxes, 2, 4].

    tran_bboxes_idxes = matmul_fn(bboxes_tran_mat, bboxes_idxes)

    # New bboxes, defined as the rectangle enclosing the transformed bboxes.
    tran_xs = tran_bboxes_idxes[:, 0, :]  # Shape: [num_bboxes, 4].
    tran_ys = tran_bboxes_idxes[:, 1, :]
    max_xs = max_fn(tran_xs, -1)  # Shape: [num_bboxes].
    max_ys = max_fn(tran_ys, -1)
    min_xs = min_fn(tran_xs, -1)
    min_ys = min_fn(tran_ys, -1)

    tran_top_left_xs = round_to_int_fn(expand_dim_fn(min_xs, -1))
    tran_top_left_ys = round_to_int_fn(expand_dim_fn(min_ys, -1))
    tran_bottom_right_xs = round_to_int_fn(expand_dim_fn(max_xs, -1))
    tran_bottom_right_ys = round_to_int_fn(expand_dim_fn(max_ys, -1))
    new_widths = tran_bottom_right_xs - tran_top_left_xs + 1
    new_heights = tran_bottom_right_ys - tran_top_left_ys + 1
    tran_bboxes = concat_fn([  # Shape: [num_bboxes, 4].
        tran_top_left_xs, tran_top_left_ys, new_widths, new_heights
    ], -1)

    new_xs = tran_bboxes[:, :1] + width // 2 + w_mod - 1
    new_ys = tran_bboxes[:, 1:2] + height // 2 + h_mod - 1

    # Filter new bboxes values.
    xcens = new_xs + new_widths // 2
    ycens = new_ys + new_heights // 2
    included = squeeze_fn(logical_and_fn(
        logical_and_fn(xcens >= 0, xcens <= width),
        logical_and_fn(ycens >= 0, ycens <= height)
    ), -1)

    # Clip new bboxes values.
    xmaxs = clip_fn(
        convert_fn(new_xs + new_widths), convert_fn(0), convert_fn(width)
    )
    ymaxs = clip_fn(
        convert_fn(new_ys + new_heights), convert_fn(0), convert_fn(height)
    )
    new_xs = clip_fn(convert_fn(new_xs), convert_fn(0), convert_fn(width))
    new_ys = clip_fn(convert_fn(new_ys), convert_fn(0), convert_fn(height))
    new_widths = xmaxs - new_xs
    new_heights = ymaxs - new_ys

    new_bboxes = concat_fn([new_xs, new_ys, new_widths, new_heights], 1)
    new_bboxes = convert_fn(boolean_mask_fn(new_bboxes, included))

    # Filter labels.
    new_labels = boolean_mask_fn(labels, included)

    return new_image, new_bboxes, new_labels


def _rotate(
        image: T,
        bboxes: T,
        labels: T,
        angle_deg: float,
        convert_fn: Callable[..., T],
        shape_fn: Callable[[T], Tuple[int, ...]],
        reshape_fn: Callable[[T, Tuple[int, ...]], T],
        expand_dim_fn: Callable[[T, int], T],
        squeeze_fn: Callable[[T, int], T],
        pad_image_fn: Callable[[T, T], T],
        range_fn: Callable[[int, int, int], T],
        round_to_int_fn: Callable[[T], T],
        repeat_fn: Callable[[T, T], T],
        tile_fn: Callable[[T, T], T],
        stack_fn: Callable[[List[T], int], T],
        concat_fn: Callable[[List[T], int], T],
        cos_fn: Callable[[T], T],
        sin_fn: Callable[[T], T],
        matmul_fn: Callable[[T, T], T],
        clip_fn: Callable[[T, T, T], T],
        gather_image_fn: Callable[[T, T], T],
        copy_fn: Callable[[T], T],
        max_fn: Callable[[T, int], T],
        min_fn: Callable[[T, int], T],
        logical_and_fn: Callable[[T, T], T],
        boolean_mask_fn: Callable[[T, T], T]
) -> Tuple[T, T, T]:
    """
    image: [h, w, c]
    bboxes: [[top_left_x, top_left_y, width, height], ...]
    labels: [0, 1, 0, ...]
    angle_deg: goes anti-clockwise.
    """
    if angle_deg == 0.0:
        return image, bboxes, labels

    ang_rad = convert_fn(np.pi * angle_deg / 180.0)

    # Image rotation matrix. Clockwise for the destination indices,
    # so the final image would appear to be rotated anti-clockwise.
    image_dest_rot_mat = convert_fn([
        [cos_fn(ang_rad), -sin_fn(ang_rad)],
        [sin_fn(ang_rad), cos_fn(ang_rad)]
    ])

    bboxes_rot_mat = convert_fn([  # Anti-clockwise.
        [cos_fn(ang_rad), sin_fn(ang_rad)],
        [-sin_fn(ang_rad), cos_fn(ang_rad)]
    ])

    return _affine_transform(
        image, bboxes, labels, convert_fn, shape_fn, reshape_fn,
        expand_dim_fn, squeeze_fn, pad_image_fn, range_fn, round_to_int_fn,
        repeat_fn, tile_fn, stack_fn, concat_fn,
        image_dest_rot_mat, bboxes_rot_mat, matmul_fn,
        clip_fn, gather_image_fn, copy_fn, max_fn, min_fn,
        logical_and_fn, boolean_mask_fn
    )


def _shear(
        image: T,
        bboxes: T,
        labels: T,
        angle_deg: float,
        convert_fn: Callable[..., T],
        shape_fn: Callable[[T], Tuple[int, ...]],
        reshape_fn: Callable[[T, Tuple[int, ...]], T],
        expand_dim_fn: Callable[[T, int], T],
        squeeze_fn: Callable[[T, int], T],
        pad_image_fn: Callable[[T, T], T],
        range_fn: Callable[[int, int, int], T],
        round_to_int_fn: Callable[[T], T],
        repeat_fn: Callable[[T, T], T],
        tile_fn: Callable[[T, T], T],
        stack_fn: Callable[[List[T], int], T],
        concat_fn: Callable[[List[T], int], T],
        tan_fn: Callable[[T], T],
        matmul_fn: Callable[[T, T], T],
        clip_fn: Callable[[T, T, T], T],
        gather_image_fn: Callable[[T, T], T],
        copy_fn: Callable[[T], T],
        max_fn: Callable[[T, int], T],
        min_fn: Callable[[T, int], T],
        logical_and_fn: Callable[[T, T], T],
        boolean_mask_fn: Callable[[T, T], T]
) -> Tuple[T, T, T]:
    """
    image: [h, w, c]
    bboxes: [[top_left_x, top_left_y, width, height], ...]
    labels: [0, 1, 0, ...]
    angle_deg: goes anti-clockwise, where abs(angle_deg) must be < 90.
    """
    if angle_deg == 0.0:
        return image, bboxes, labels

    ang_rad = convert_fn(np.pi * angle_deg / 180.0)
    factor = tan_fn(ang_rad)

    # Image shear matrix. Clockwise for the destination indices,
    # so the final image would appear to be sheared anti-clockwise.
    image_dest_shear_mat = convert_fn([
        [convert_fn(1), -factor],
        [convert_fn(0), convert_fn(1)]
    ])

    bboxes_shear_mat = convert_fn([  # Anti-clockwise.
        [convert_fn(1), factor],
        [convert_fn(0), convert_fn(1)]
    ])

    return _affine_transform(
        image, bboxes, labels, convert_fn, shape_fn, reshape_fn,
        expand_dim_fn, squeeze_fn, pad_image_fn, range_fn, round_to_int_fn,
        repeat_fn, tile_fn, stack_fn, concat_fn,
        image_dest_shear_mat, bboxes_shear_mat, matmul_fn,
        clip_fn, gather_image_fn, copy_fn, max_fn, min_fn,
        logical_and_fn, boolean_mask_fn
    )


def _resize(
        image: T,
        bboxes: T,
        labels: T,
        dest_size: Tuple[int, int],
        convert_fn: Callable[..., T],
        shape_fn: Callable[[T], Tuple[int, ...]],
        reshape_fn: Callable[[T, Tuple[int, ...]], T],
        resize_image_fn: Callable[[T, Tuple[int, int]], T],
        concat_fn: Callable[[List[T], int], T],
) -> Tuple[T, T, T]:
    """
    image: [h, w, c]
    bboxes: [[top_left_x, top_left_y, width, height], ...]
    labels: [0, 1, 0, ...]
    dest_size: (height, width)
    """
    image, bboxes, labels = _sanitise(
        image, bboxes, labels, convert_fn, reshape_fn
    )

    image_shape = shape_fn(image)
    assert len(image_shape) == 3

    if image_shape[0] == dest_size[0] and image_shape[1] == dest_size[1]:
        return image, bboxes, labels

    image = resize_image_fn(image, dest_size)

    w = convert_fn(dest_size[1] / image_shape[1])
    h = convert_fn(dest_size[0] / image_shape[0])

    xs = bboxes[:, :1] * w
    ys = bboxes[:, 1:2] * h
    widths = bboxes[:, 2:3] * w
    heights = bboxes[:, 3:] * h
    bboxes = concat_fn([xs, ys, widths, heights], 1)

    return image, bboxes, labels


def _get_random_size_fractions(
        height_fraction_range: Tuple[float, float],
        width_fraction_range: Tuple[float, float],
        rand_fn: Callable[..., T],
        convert_fn: Callable[..., T],
) -> Tuple[T, T]:
    """
    height_fraction_range, width_fraction_range: (-1.0, 1.0)
    rand_fn: generate array of random numbers in range [0.0, 1.0)
    """
    height_fraction_range = convert_fn(height_fraction_range)
    width_fraction_range = convert_fn(width_fraction_range)

    min_height_fraction = height_fraction_range[0]
    min_width_fraction = width_fraction_range[0]
    height_fraction_diff = height_fraction_range[1] - min_height_fraction
    width_fraction_diff = width_fraction_range[1] - min_width_fraction

    height_fractions = height_fraction_diff * rand_fn() + min_height_fraction
    width_fractions = width_fraction_diff * rand_fn() + min_width_fraction

    return height_fractions, width_fractions


def _get_random_crop_inputs(
        image_height: int,
        image_width: int,
        height_fraction_range: Tuple[float, float],
        width_fraction_range: Tuple[float, float],
        rand_fn: Callable[..., T],
        convert_fn: Callable[..., T],
        round_to_int_fn: Callable[[T], T]
) -> Tuple[T, T, T, T]:
    """
    height_fraction_range, width_fraction_range: in range [0.0, 1.0)
    rand_fn: generate array of random numbers in range [0.0, 1.0)
    Return: randomized (offset_heights, offset_widths,
                        cropped_image_heights, cropped_image_widths)
    """
    image_height = convert_fn(image_height)
    image_width = convert_fn(image_width)
    height_fraction_range = convert_fn(height_fraction_range)
    width_fraction_range = convert_fn(width_fraction_range)

    height_fractions, width_fractions = _get_random_size_fractions(
        height_fraction_range, width_fraction_range, rand_fn, convert_fn
    )

    cropped_image_heights = image_height * height_fractions
    cropped_image_widths = image_height * width_fractions

    offset_heights = round_to_int_fn(
        (image_height - cropped_image_heights) * rand_fn()
    )
    offset_widths = round_to_int_fn(
        (image_width - cropped_image_widths) * rand_fn()
    )

    return (
        offset_heights, offset_widths,
        cropped_image_heights, cropped_image_widths
    )


def _crop(
        image: T,
        bboxes: T,
        labels: T,
        offset_height: int,
        offset_width: int,
        cropped_image_height: int,
        cropped_image_width: int,
        convert_fn: Callable[..., T],
        shape_fn: Callable[[T], Tuple[int, ...]],
        reshape_fn: Callable[[T, Tuple[int, ...]], T],
        concat_fn: Callable[[List[T], int], T],
        logical_and_fn: Callable[[T, T], T],
        squeeze_fn: Callable[[T, int], T],
        clip_fn: Callable[[T, T, T], T],
        boolean_mask_fn: Callable[[T, T], T],
) -> Tuple[T, T, T]:
    """
    image: [h, w, c]
    bboxes: [[top_left_x, top_left_y, width, height], ...]
    labels: [0, 1, 0, ...]
    offset_height: in range [0, image_height - 1]
    offset_width: in range [0, image_width - 1]
    """
    image, bboxes, labels = _sanitise(
        image, bboxes, labels, convert_fn, reshape_fn
    )

    image_shape = shape_fn(image)
    assert len(image_shape) == 3

    offset_height = convert_fn(offset_height)
    offset_width = convert_fn(offset_width)
    cropped_image_height = convert_fn(cropped_image_height)
    cropped_image_width = convert_fn(cropped_image_width)

    top = offset_height
    left = offset_width
    bottom = top + cropped_image_height
    right = left + cropped_image_width

    # Crop image.
    image = image[int(top):int(bottom), int(left):int(right), :]

    # Translate bboxes values.
    xs = bboxes[:, :1] - offset_width
    ys = bboxes[:, 1:2] - offset_height
    widths = bboxes[:, 2:3]
    heights = bboxes[:, 3:]

    # Filter bboxes values.
    xcens = xs + widths // 2
    ycens = ys + heights // 2
    included = squeeze_fn(logical_and_fn(
        logical_and_fn(xcens >= 0, xcens <= cropped_image_width),
        logical_and_fn(ycens >= 0, ycens <= cropped_image_height)
    ), -1)  # Squeeze along the last axis.

    # Clip bboxes values.
    xmaxs = clip_fn(xs + widths, convert_fn(0), cropped_image_width)
    ymaxs = clip_fn(ys + heights, convert_fn(0), cropped_image_height)
    xs = clip_fn(xs, convert_fn(0), cropped_image_width)
    ys = clip_fn(ys, convert_fn(0), cropped_image_height)
    widths = xmaxs - xs
    heights = ymaxs - ys

    bboxes = concat_fn([xs, ys, widths, heights], 1)
    bboxes = boolean_mask_fn(bboxes, included)

    # Filter labels.
    labels = boolean_mask_fn(labels, included)

    return image, bboxes, labels


def _translate(
        image: T,
        bboxes: T,
        labels: T,
        translate_height: int,
        translate_width: int,
        convert_fn: Callable[..., T],
        shape_fn: Callable[[T], Tuple[int, ...]],
        reshape_fn: Callable[[T, Tuple[int, ...]], T],
        where_fn: Callable[[T, T, T], T],
        abs_fn: Callable[[T], T],
        concat_fn: Callable[[List[T], int], T],
        logical_and_fn: Callable[[T, T], T],
        squeeze_fn: Callable[[T, int], T],
        clip_fn: Callable[[T, T, T], T],
        boolean_mask_fn: Callable[[T, T], T],
        pad_image_fn: Callable[[T, T], T],
) -> Tuple[T, T, T]:
    """
    Making use of cropping and padding to perform translation.

    image: [h, w, c]
    bboxes: [[top_left_x, top_left_y, width, height], ...]
    labels: [0, 1, 0, ...]
    translate_height: in range [-image_height + 1: image_height - 1]
    translate_width: in range [-image_width + 1: image_width - 1]
    """
    image, bboxes, labels = _sanitise(
        image, bboxes, labels, convert_fn, reshape_fn
    )

    image_shape = shape_fn(image)
    assert len(image_shape) == 3

    translate_height = convert_fn(translate_height)
    translate_width = convert_fn(translate_width)

    t = where_fn(
        translate_height >= 0,
        convert_fn([0, translate_height, 0]),
        convert_fn([-translate_height, 0, -translate_height])
    )
    offset_height, pad_top, pad_bottom = t[0], t[1], t[2]

    t = where_fn(
        translate_width >= 0,
        convert_fn([0, translate_width, 0]),
        convert_fn([-translate_width, 0, -translate_width])
    )
    offset_width, pad_left, pad_right = t[0], t[1], t[2]

    cropped_height = convert_fn(image_shape[0]) - abs_fn(translate_height)
    cropped_width = convert_fn(image_shape[1]) - abs_fn(translate_width)

    image, bboxes, labels = _crop(
        image, bboxes, labels,
        offset_height, offset_width, cropped_height, cropped_width,
        convert_fn, shape_fn, reshape_fn, concat_fn, logical_and_fn,
        squeeze_fn, clip_fn, boolean_mask_fn
    )

    image = pad_image_fn(
        image, convert_fn([pad_top, pad_bottom, pad_left, pad_right])
    )

    bboxes = _translate_bboxes(bboxes, pad_top, pad_left, concat_fn)

    return image, bboxes, labels



"""
Unit tests.
"""

import unittest

import numpy as np  # type: ignore
import tensorflow as tf  # type: ignore

from .np import (
    flip_left_right,
    flip_up_down,
    rotate_90,
    rotate_90_and_pad,
    rotate,
    crop,
    translate,
    shear
)

from .tf import (
    np_to_tf,
    tf_flip_left_right,
    tf_flip_up_down,
    tf_rotate_90,
    tf_rotate_90_and_pad,
    tf_rotate,
    tf_crop,
    tf_translate,
    tf_shear
)


ORIGINAL_IMAGE_LIST = [
    np.array([
        [[1], [2], [3]],
        [[4], [5], [6]],
        [[7], [8], [9]],
        [[10], [11], [12]]
    ], dtype=np.float32),
    np.array([
        [[11], [12], [13]],
        [[14], [15], [16]],
        [[17], [18], [19]],
        [[20], [21], [22]]
    ], dtype=np.float32),
    np.array([
        [[21], [22], [23]],
        [[24], [25], [26]],
        [[27], [28], [29]],
        [[30], [31], [32]]
    ], dtype=np.float32),
]

ORIGINAL_BBOXES_LIST = [
    np.array([
        [1, 0, 2, 2],
        [0, 1, 3, 2],
    ], dtype=np.float32),
    np.array([
        [0, 0, 2, 3],
    ], dtype=np.float32),
    np.array([], dtype=np.float32).reshape(-1, 4),
]

ORIGINAL_LABELS_LIST = [
    np.array([0, 1], dtype=np.float32),
    np.array([2], dtype=np.float32),
    np.array([], dtype=np.float32),
]


(
    TF_ORIGINAL_IMAGE_LIST, TF_ORIGINAL_BBOXES_LIST, TF_ORIGINAL_LABELS_LIST
) = np_to_tf(
    ORIGINAL_IMAGE_LIST, ORIGINAL_BBOXES_LIST, ORIGINAL_LABELS_LIST
)


class TestTransform(unittest.TestCase):

    def test_flip_left_right(self) -> None:

        expected_image_list = [
            np.array([
                [[3], [2], [1]],
                [[6], [5], [4]],
                [[9], [8], [7]],
                [[12], [11], [10]]
            ], dtype=np.float32),
            np.array([
                [[13], [12], [11]],
                [[16], [15], [14]],
                [[19], [18], [17]],
                [[22], [21], [20]]
            ], dtype=np.float32),
            np.array([
                [[23], [22], [21]],
                [[26], [25], [24]],
                [[29], [28], [27]],
                [[32], [31], [30]]
            ], dtype=np.float32),
        ]
        expected_bboxes_list = [
            np.array([
                [0, 0, 2, 2],
                [0, 1, 3, 2],
            ], dtype=np.float32),
            np.array([
                [1, 0, 2, 3],
            ], dtype=np.float32),
            np.array([], dtype=np.float32).reshape(-1, 4),
        ]
        expected_labels_list = ORIGINAL_LABELS_LIST

        # Numpy.
        for i in range(len(ORIGINAL_IMAGE_LIST)):
            image, bboxes, labels = flip_left_right(
                ORIGINAL_IMAGE_LIST[i],
                ORIGINAL_BBOXES_LIST[i],
                ORIGINAL_LABELS_LIST[i]
            )
            self.assertTrue(
                np.array_equal(expected_image_list[i], image)
            )
            self.assertTrue(
                np.array_equal(expected_bboxes_list[i], bboxes)
            )
            self.assertTrue(
                np.array_equal(expected_labels_list[i], labels)
            )

        # TF.
        (
            tf_expected_image_list,
            tf_expected_bboxes_list,
            tf_expected_labels_list
        ) = np_to_tf(
            expected_image_list, expected_bboxes_list, expected_labels_list
        )
        for i in range(len(TF_ORIGINAL_LABELS_LIST)):
            tf_image, tf_bboxes, tf_labels = tf_flip_left_right(
                TF_ORIGINAL_IMAGE_LIST[i],
                TF_ORIGINAL_BBOXES_LIST[i],
                TF_ORIGINAL_LABELS_LIST[i]
            )
            self.assertTrue(
                np.array_equal(tf_expected_image_list[i].numpy(),
                               tf_image.numpy())
            )
            self.assertTrue(
                np.array_equal(tf_expected_bboxes_list[i].numpy(),
                               tf_bboxes.numpy())
            )
            self.assertTrue(
                np.array_equal(tf_expected_labels_list[i].numpy(),
                               tf_labels.numpy())
            )

    def test_flip_up_down(self) -> None:

        expected_image_list = [
            np.array([
                [[10], [11], [12]],
                [[7], [8], [9]],
                [[4], [5], [6]],
                [[1], [2], [3]]
            ], dtype=np.float32),
            np.array([
                [[20], [21], [22]],
                [[17], [18], [19]],
                [[14], [15], [16]],
                [[11], [12], [13]]
            ], dtype=np.float32),
            np.array([
                [[30], [31], [32]],
                [[27], [28], [29]],
                [[24], [25], [26]],
                [[21], [22], [23]]
            ], dtype=np.float32),
        ]
        expected_bboxes_list = [
            np.array([
                [1, 2, 2, 2],
                [0, 1, 3, 2],
            ], dtype=np.float32),
            np.array([
                [0, 1, 2, 3],
            ], dtype=np.float32),
            np.array([], dtype=np.float32).reshape(-1, 4),
        ]
        expected_labels_list = ORIGINAL_LABELS_LIST

        # Numpy.
        for i in range(len(ORIGINAL_IMAGE_LIST)):
            image, bboxes, labels = flip_up_down(
                ORIGINAL_IMAGE_LIST[i],
                ORIGINAL_BBOXES_LIST[i],
                ORIGINAL_LABELS_LIST[i]
            )
            self.assertTrue(
                np.array_equal(expected_image_list[i], image)
            )
            self.assertTrue(
                np.array_equal(expected_bboxes_list[i], bboxes)
            )
            self.assertTrue(
                np.array_equal(expected_labels_list[i], labels)
            )

        # TF.
        (
            tf_expected_image_list,
            tf_expected_bboxes_list,
            tf_expected_labels_list
        ) = np_to_tf(
            expected_image_list, expected_bboxes_list, expected_labels_list
        )
        for i in range(len(TF_ORIGINAL_LABELS_LIST)):
            tf_image, tf_bboxes, tf_labels = tf_flip_up_down(
                TF_ORIGINAL_IMAGE_LIST[i],
                TF_ORIGINAL_BBOXES_LIST[i],
                TF_ORIGINAL_LABELS_LIST[i]
            )
            self.assertTrue(
                np.array_equal(tf_expected_image_list[i].numpy(),
                               tf_image.numpy())
            )
            self.assertTrue(
                np.array_equal(tf_expected_bboxes_list[i].numpy(),
                               tf_bboxes.numpy())
            )
            self.assertTrue(
                np.array_equal(tf_expected_labels_list[i].numpy(),
                               tf_labels.numpy())
            )

    def test_rotate_90(self) -> None:

        expected_image_list = [
            np.array([
                [[3], [6], [9], [12]],
                [[2], [5], [8], [11]],
                [[1], [4], [7], [10]]
            ], dtype=np.float32),
            np.array([
                [[13], [16], [19], [22]],
                [[12], [15], [18], [21]],
                [[11], [14], [17], [20]]
            ], dtype=np.float32),
            np.array([
                [[23], [26], [29], [32]],
                [[22], [25], [28], [31]],
                [[21], [24], [27], [30]]
            ], dtype=np.float32),
        ]
        expected_bboxes_list = [
            np.array([
                [0, 0, 2, 2],
                [1, 0, 2, 3],
            ], dtype=np.float32),
            np.array([
                [0, 1, 3, 2],
            ], dtype=np.float32),
            np.array([], dtype=np.float32).reshape(-1, 4),
        ]
        expected_labels_list = ORIGINAL_LABELS_LIST

        # Numpy.
        for i in range(len(ORIGINAL_IMAGE_LIST)):
            image, bboxes, labels = rotate_90(
                ORIGINAL_IMAGE_LIST[i],
                ORIGINAL_BBOXES_LIST[i],
                ORIGINAL_LABELS_LIST[i]
            )
            self.assertTrue(
                np.array_equal(expected_image_list[i], image)
            )
            self.assertTrue(
                np.array_equal(expected_bboxes_list[i], bboxes)
            )
            self.assertTrue(
                np.array_equal(expected_labels_list[i], labels)
            )

        # TF.
        (
            tf_expected_image_list,
            tf_expected_bboxes_list,
            tf_expected_labels_list
        ) = np_to_tf(
            expected_image_list, expected_bboxes_list, expected_labels_list
        )
        for i in range(len(TF_ORIGINAL_LABELS_LIST)):
            tf_image, tf_bboxes, tf_labels = tf_rotate_90(
                TF_ORIGINAL_IMAGE_LIST[i],
                TF_ORIGINAL_BBOXES_LIST[i],
                TF_ORIGINAL_LABELS_LIST[i]
            )
            self.assertTrue(
                np.array_equal(tf_expected_image_list[i].numpy(),
                               tf_image.numpy())
            )
            self.assertTrue(
                np.array_equal(tf_expected_bboxes_list[i].numpy(),
                               tf_bboxes.numpy())
            )
            self.assertTrue(
                np.array_equal(tf_expected_labels_list[i].numpy(),
                               tf_labels.numpy())
            )

    def test_rotate_90_and_pad(self) -> None:

        expected_image_list = [
            np.array([
                [[0], [0], [0], [0]],
                [[0], [0], [0], [0]],
                [[3], [6], [9], [12]],
                [[2], [5], [8], [11]],
                [[1], [4], [7], [10]],
                [[0], [0], [0], [0]]
            ], dtype=np.float32),
            np.array([
                [[0], [0], [0], [0]],
                [[0], [0], [0], [0]],
                [[13], [16], [19], [22]],
                [[12], [15], [18], [21]],
                [[11], [14], [17], [20]],
                [[0], [0], [0], [0]]
            ], dtype=np.float32),
            np.array([
                [[0], [0], [0], [0]],
                [[0], [0], [0], [0]],
                [[23], [26], [29], [32]],
                [[22], [25], [28], [31]],
                [[21], [24], [27], [30]],
                [[0], [0], [0], [0]]
            ], dtype=np.float32),
        ]
        expected_bboxes_list = [
            np.array([
                [0, 2, 2, 2],
                [1, 2, 2, 3],
            ], dtype=np.float32),
            np.array([
                [0, 3, 3, 2],
            ], dtype=np.float32),
            np.array([], dtype=np.float32).reshape(-1, 4),
        ]
        expected_labels_list = ORIGINAL_LABELS_LIST

        # Numpy.
        for i in range(len(ORIGINAL_IMAGE_LIST)):
            image, bboxes, labels = rotate_90_and_pad(
                ORIGINAL_IMAGE_LIST[i],
                ORIGINAL_BBOXES_LIST[i],
                ORIGINAL_LABELS_LIST[i]
            )
            self.assertTrue(
                np.array_equal(expected_image_list[i], image)
            )
            self.assertTrue(
                np.array_equal(expected_bboxes_list[i], bboxes)
            )
            self.assertTrue(
                np.array_equal(expected_labels_list[i], labels)
            )

        # TF.
        (
            tf_expected_image_list,
            tf_expected_bboxes_list,
            tf_expected_labels_list
        ) = np_to_tf(
            expected_image_list, expected_bboxes_list, expected_labels_list
        )
        for i in range(len(TF_ORIGINAL_LABELS_LIST)):
            tf_image, tf_bboxes, tf_labels = tf_rotate_90_and_pad(
                TF_ORIGINAL_IMAGE_LIST[i],
                TF_ORIGINAL_BBOXES_LIST[i],
                TF_ORIGINAL_LABELS_LIST[i]
            )
            self.assertTrue(
                np.array_equal(tf_expected_image_list[i].numpy(),
                               tf_image.numpy())
            )
            self.assertTrue(
                np.array_equal(tf_expected_bboxes_list[i].numpy(),
                               tf_bboxes.numpy())
            )
            self.assertTrue(
                np.array_equal(tf_expected_labels_list[i].numpy(),
                               tf_labels.numpy())
            )

    def test_rotate(self) -> None:

        original_image_list = [
            np.array([
                [[1], [2], [3], [4]],
                [[5], [6], [7], [8]],
                [[9], [10], [11], [12]]
            ], dtype=np.float32),
            np.array([
                [[13], [14], [15], [16]],
                [[17], [18], [19], [20]],
                [[21], [22], [23], [24]]
            ], dtype=np.float32),
        ]
        original_bboxes_list = [
            np.array([
                [1, 0, 2, 2],
                [0, 1, 3, 2],
            ], dtype=np.float32),
            np.array([], dtype=np.float32).reshape(-1, 4),
        ]
        original_labels_list = [
            np.array([1, 2], dtype=np.float32),
            np.array([], dtype=np.float32),
        ]

        angles_deg = [90.0, 180.0]

        expected_image_list = [
            np.array([
                [[3], [7], [11], [0]],
                [[2], [6], [10], [0]],
                [[1], [5], [9], [0]]
            ], dtype=np.float32),
            np.array([
                [[23], [22], [21], [0]],
                [[19], [18], [17], [0]],
                [[15], [14], [13], [0]]
            ], dtype=np.float32),
        ]
        expected_bboxes_list = [
            np.array([
                [0, 0, 2, 2],
                [1, 0, 2, 3],
            ], dtype=np.float32),
            np.array([], dtype=np.float32).reshape(-1, 4),
        ]
        expected_labels_list = original_labels_list

        # Numpy.
        for i in range(len(original_image_list)):
            image, bboxes, labels = rotate(
                original_image_list[i],
                original_bboxes_list[i],
                original_labels_list[i],
                angles_deg[i]
            )
            self.assertTrue(
                np.array_equal(expected_image_list[i], image)
            )
            self.assertTrue(
                np.array_equal(expected_bboxes_list[i], bboxes)
            )
            self.assertTrue(
                np.array_equal(expected_labels_list[i], labels)
            )

        # TF.
        (
            tf_original_image_list,
            tf_original_bboxes_list,
            tf_original_labels_list
        ) = np_to_tf(
            original_image_list, original_bboxes_list, original_labels_list
        )
        (
            tf_expected_image_list,
            tf_expected_bboxes_list,
            tf_expected_labels_list
        ) = np_to_tf(
            expected_image_list, expected_bboxes_list, expected_labels_list
        )
        for i in range(len(tf_original_image_list)):
            tf_image, tf_bboxes, tf_labels = tf_rotate(
                tf_original_image_list[i],
                tf_original_bboxes_list[i],
                tf_original_labels_list[i],
                angles_deg[i]
            )
            self.assertTrue(
                np.array_equal(tf_expected_image_list[i].numpy(),
                               tf_image.numpy())
            )
            self.assertTrue(
                np.array_equal(tf_expected_bboxes_list[i].numpy(),
                               tf_bboxes.numpy())
            )
            self.assertTrue(
                np.array_equal(tf_expected_labels_list[i].numpy(),
                               tf_labels.numpy())
            )

    def test_shear(self) -> None:

        dummy_image = np.random.rand(32, 32, 3)
        original_bboxes = np.array([
            [15, 15, 2, 2],
            [15, 0, 2, 2],
        ], dtype=np.float32)
        original_labels = np.array([[0], [1]], dtype=np.float32)

        angle_deg = 45.0

        expected_bboxes = np.array([
            [15, 15, 3, 2],
            [0, 0, 3, 2],
        ], dtype=np.float32)
        expected_labels = original_labels

        # Numpy.
        _, bboxes, labels = shear(
            dummy_image, original_bboxes, original_labels,
            angle_deg
        )
        self.assertTrue(np.array_equal(expected_bboxes, bboxes))
        self.assertTrue(np.array_equal(expected_labels, labels))

        # TF.
        tf_expected_bboxes = tf.convert_to_tensor(
            expected_bboxes, dtype=tf.float32
        )
        tf_expected_labels = tf.convert_to_tensor(
            expected_labels, dtype=tf.float32
        )

        _, tf_bboxes, tf_labels = tf_shear(
            tf.convert_to_tensor(dummy_image, dtype=tf.float32),
            tf.convert_to_tensor(original_bboxes, dtype=tf.float32),
            tf.convert_to_tensor(original_labels, dtype=tf.float32),
            angle_deg
        )
        self.assertTrue(
            np.allclose(tf_expected_bboxes.numpy(), tf_bboxes.numpy())
        )
        self.assertTrue(
            np.allclose(tf_expected_labels.numpy(), tf_labels.numpy())
        )

    def test_crop(self) -> None:

        dummy_image_list = [np.random.rand(128, 128, 3) for _ in range(4)]
        original_bboxes_list = [
            np.array([
                [64, 52, 20, 24],
                [44, 48, 12, 8],
            ], dtype=np.float32),
            np.array([
                [24, 12, 20, 24],
                [108, 120, 12, 8],
            ], dtype=np.float32),
            np.array([
                [108, 120, 12, 8],
            ], dtype=np.float32),
            np.array([], dtype=np.float32).reshape(-1, 4),
        ]
        original_labels_list = [
            np.array([0, 1], dtype=np.float32),
            np.array([2, 1], dtype=np.float32),
            np.array([0], dtype=np.float32),
            np.array([], dtype=np.float32),
        ]

        offset_heights = [128 * 0.25, 0.0, 0.0, 99.0]
        offset_widths = [128 * 0.25, 0.0, 0.0, 59.0]
        cropped_image_heights = [128 * 0.75] * 4
        cropped_image_widths = [128 * 0.75] * 4

        expected_bboxes_list = [
            np.array([
                [32, 20, 20, 24],
                [12, 16, 12, 8],
            ], dtype=np.float32),
            np.array([
                [24, 12, 20, 24],
            ], dtype=np.float32),
            np.array([], dtype=np.float32).reshape(-1, 4),
            np.array([], dtype=np.float32).reshape(-1, 4),
        ]
        expected_labels_list = [
            np.array([0, 1], dtype=np.float32),
            np.array([2], dtype=np.float32),
            np.array([], dtype=np.float32),
            np.array([], dtype=np.float32),
        ]

        # Numpy.
        for i in range(len(dummy_image_list)):
            _, bboxes, labels = crop(
                dummy_image_list[i],
                original_bboxes_list[i],
                original_labels_list[i],
                int(offset_heights[i]), int(offset_widths[i]),
                int(cropped_image_heights[i]), int(cropped_image_widths[i])
            )
            self.assertTrue(
                np.allclose(expected_bboxes_list[i], bboxes)
            )
            self.assertTrue(
                np.allclose(expected_labels_list[i], labels)
            )

        # TF.
        (
            tf_dummy_image_list,
            tf_original_bboxes_list,
            tf_original_labels_list
        ) = np_to_tf(
            dummy_image_list, original_bboxes_list, original_labels_list
        )
        _, tf_expected_bboxes_list, tf_expected_labels_list = np_to_tf(
            dummy_image_list, expected_bboxes_list, expected_labels_list
        )

        for i in range(len(tf_dummy_image_list)):
            _, tf_bboxes, tf_labels = tf_crop(
                tf_dummy_image_list[i],
                tf_original_bboxes_list[i],
                tf_original_labels_list[i],
                int(offset_heights[i]), int(offset_widths[i]),
                int(cropped_image_heights[i]), int(cropped_image_widths[i])
            )
            self.assertTrue(
                np.allclose(tf_expected_bboxes_list[i].numpy(),
                            tf_bboxes.numpy())
            )
            self.assertTrue(
                np.allclose(tf_expected_labels_list[i].numpy(),
                            tf_labels.numpy())
            )

    def test_translate(self) -> None:

        translate_heights = [-1, 0, 1]
        translate_widths = [0, 1, 1]

        expected_image_list = [
            np.array([
                [[4], [5], [6]],
                [[7], [8], [9]],
                [[10], [11], [12]],
                [[0], [0], [0]]
            ], dtype=np.float32),
            np.array([
                [[0], [11], [12]],
                [[0], [14], [15]],
                [[0], [17], [18]],
                [[0], [20], [21]]
            ], dtype=np.float32),
            np.array([
                [[0], [0], [0]],
                [[0], [21], [22]],
                [[0], [24], [25]],
                [[0], [27], [28]]
            ], dtype=np.float32),
        ]
        expected_bboxes_list = [
            np.array([
                [1, 0, 2, 1],
                [0, 0, 3, 2],
            ], dtype=np.float32),
            np.array([
                [1, 0, 2, 3],
            ], dtype=np.float32),
            np.array([], dtype=np.float32).reshape(-1, 4),
        ]
        expected_labels_list = [
            np.array([0, 1], dtype=np.float32),
            np.array([2], dtype=np.float32),
            np.array([], dtype=np.float32),
        ]

        # Numpy.
        for i in range(len(ORIGINAL_IMAGE_LIST)):
            image, bboxes, labels = translate(
                ORIGINAL_IMAGE_LIST[i],
                ORIGINAL_BBOXES_LIST[i],
                ORIGINAL_LABELS_LIST[i],
                translate_heights[i], translate_widths[i]
            )
            self.assertTrue(np.array_equal(expected_image_list[i], image))
            self.assertTrue(np.array_equal(expected_bboxes_list[i], bboxes))
            self.assertTrue(np.array_equal(expected_labels_list[i], labels))

        # TF.
        (
            tf_expected_image_list,
            tf_expected_bboxes_list,
            tf_expected_labels_list
        ) = np_to_tf(
            expected_image_list, expected_bboxes_list, expected_labels_list
        )

        for i in range(len(TF_ORIGINAL_LABELS_LIST)):
            tf_image, tf_bboxes, tf_labels = tf_translate(
                TF_ORIGINAL_IMAGE_LIST[i],
                TF_ORIGINAL_BBOXES_LIST[i],
                TF_ORIGINAL_LABELS_LIST[i],
                translate_heights[i], translate_widths[i]
            )
            self.assertTrue(
                np.array_equal(tf_expected_image_list[i].numpy(),
                               tf_image.numpy())
            )
            self.assertTrue(
                np.array_equal(tf_expected_bboxes_list[i].numpy(),
                               tf_bboxes.numpy())
            )
            self.assertTrue(
                np.array_equal(tf_expected_labels_list[i].numpy(),
                               tf_labels.numpy())
            )


if __name__ == "__main__":
    unittest.main()



"""
Utilities.
"""

from typing import Any, Callable, Sequence, Tuple, TypeVar

T = TypeVar("T")


class Compose:

    def __init__(self, fns: Sequence[Callable[..., Any]]) -> None:
        self.fns = fns

    def __call__(self, *args: Any) -> Any:
        for fn in self.fns:
            args = fn(*args)
        return args


def collate_fn(batch_: Sequence[Tuple[Any, ...]]) -> Tuple[Sequence[Any], ...]:
    return tuple(zip(*batch_))


def image_only(
        tran_fn: Callable[[T, Any, Any], Tuple[T, T, T]]
) -> Callable[[T, T], Tuple[T, T]]:

    def fn(image: T, *args: Any) -> Any:
        transformed_image = tran_fn(image, [], [])[0]
        if not len(args):
            return transformed_image
        return (transformed_image, *args)

    return fn



from ._np import (
    flip_left_right,
    flip_up_down,
    rotate_90,
    rotate_90_and_pad,
    rotate,
    crop,
    translate,
    shear,
    Resize,
    RandomTransform,
    RandomFlipLeftRight,
    RandomFlipUpDown,
    RandomRotate90,
    RandomRotate90AndPad,
    RandomRotate,
    RandomCrop,
    RandomTranslate,
    RandomShear
)
